<li class="navSearchWrapper">
  <input id="{{ include.inputselector }}" type="search" />
</li>
<script type="text/javascript" src="https://cdn.jsdelivr.net/docsearch.js/1/docsearch.min.js"></script>
<script>
// For Algolia search
(function() {
  // Algolia
  docsearch({
    apiKey: '{{ site.searchconfig.apikey }}',
    indexName: '{{ site.searchconfig.indexname }}',
    inputSelector: '#{{ include.inputselector }}',
  });
}());
</script>