<div class="headerContainer">
  <div id="header_wrap" class="wrapper headerWrapper">
    <div class="inner">
      <img class="projectLogo" height="200px" src="{{ '/static/logo.svg' }}" alt="{{ site.title }}" title="{{ site.title }}" />
      <h1 id="project_title">{{ site.title }}</h1>
      <h2 id="project_tagline" class="fbossFontLight">{{ site.tagline }}</h2>

      <section id="intro">
        <p>{% if page.excerpt %}{{ page.excerpt | strip_html }}{% else %}{{ site.description }}{% endif %}</p>
      </section>
      <div id="promo" class="section promoSection">
        {% for promo in site.data.promo %}
          {% include plugins/{{promo.type}}.html button_href=promo.href button_text=promo.text %}
          <div class="gridClear"></div>
        {% endfor %}
      </div>
    </div>
  </div>
</div>
