{% if include.layout == "fourColumn" %}
  {% assign layout = "fourByGridBlock" %}
{% else %}
  {% assign layout = "twoByGridBlock" %}
{% endif %}

{% if include.imagealign == "side" %}
  {% assign imagealign = "imageAlignSide" %}
{% else %}
  {% if item.image %}
    {% assign imagealign = "imageAlignTop" %}
  {% else %}
    {% assign imagealign = "" %}
  {% endif %}
{% endif %}

{% if include.align == "right" %}
  {% assign align = "alignRight" %}
{% elsif include.align == "center" %}
  {% assign align = "alignCenter" %}
{% else %}
  {% assign align = "alignLeft" %}
{% endif %}

<div class="blockElement {{ layout }} {{ imagealign }} {{ align }}">
  {% if item.image %}
  <div class="blockImage">
    <img src="/static/{{ item.image }}" alt="{{ item.title }}" title="{{ item.title }}" />
  </div>
  {% endif %}
  <div class="blockContent">
    <h3>{{ item.title }}</h3>
    {% if item.text %}
    {{ item.text | markdownify }}
    {% endif %}
  </div>
</div>
