<!DOCTYPE html>
<html>
  {% include head.html %}
  <body>
    {% include nav.html alwayson=true %}
    <div class="navPusher">
      {% include home_header.html %}
      <div class="mainContainer">
        <div id="main_wrap" class="wrapper mainWrapper">
          {{ content }}
        </div>
        {% include powered_by.html %}
      </div>
      {% include footer.html %}
    </div>
  </body>
</html>
