<div class="gridBlock">
{% for item in {{include.data_source}} %}
  {% include content/items/gridblock.html item=item layout=include.layout imagealign=include.imagealign align=include.align %}
{% endfor %}
</div>