DROP USER mtr CASCADE;
CREATE USER mtr IDENTIFIED BY mtr
DEFAULT TABLESPACE users
TEMPORARY TABLESPACE temp;
GRANT CREATE SESSION TO mtr;
GRANT CREATE TABLE TO mtr;
GRANT CREATE VIEW TO mtr;
ALTER USER mtr QUOTA UNLIMITED ON USERS;
CONNECT mtr/mtr;
CREATE TABLE t1 (a INT,b NUMBER);
INSERT INTO t1 VALUES (10,1000*1000*1000);
INSERT INTO t1 VALUES (20,1000*1000*1000*1000);
INSERT INTO t1 VALUES (30,1000*1000*1000*1000*1000);
CREATE VIEW v1 AS SELECT * FROM t1;
CREATE TABLE t2 (a VARCHAR(64));
INSERT INTO t2 VALUES ('test');
