#
--disable_query_log
--eval LOAD DATA LOCAL INFILE '$MYSQL_TMP_DIR/region.tbl' INTO TABLE region FIELDS TERMINATED BY '|'
--eval LOAD DATA LOCAL INFILE '$MYSQL_TMP_DIR/nation.tbl' INTO TABLE nation FIELDS TERMINATED BY '|'
--eval LOAD DATA LOCAL INFILE '$MYSQL_TMP_DIR/customer.tbl' INTO TABLE customer FIELDS TERMINATED BY '|'
--eval LOAD DATA LOCAL INFILE '$MYSQL_TMP_DIR/supplier.tbl' INTO TABLE supplier FIELDS TERMINATED BY '|'
--eval LOAD DATA LOCAL INFILE '$MYSQL_TMP_DIR/part.tbl' INTO TABLE part FIELDS TERMINATED BY '|'
--eval LOAD DATA LOCAL INFILE '$MYSQL_TMP_DIR/partsupp.tbl' INTO TABLE partsupp FIELDS TERMINATED BY '|'
--eval LOAD DATA LOCAL INFILE '$MYSQL_TMP_DIR/orders.tbl' INTO TABLE orders FIELDS TERMINATED BY '|'
--eval LOAD DATA LOCAL INFILE '$MYSQL_TMP_DIR/lineitem.tbl' INTO TABLE lineitem FIELDS TERMINATED BY '|'
--enable_query_log
#
