drop table if exists datatypeupdate;
Create table datatypeupdate (
CIDX 		INTEGER,
CBIGINT 	BIGINT,
CDECIMAL1 	DECIMAL(1),
CDECIMAL4 	DECIMAL(4),
CDECIMAL4_2 	DECIMAL(4,2),
CDECIMAL5 	DECIMAL(5),
CDECIMAL9 	DECIMAL(9),
CDECIMAL9_2 	DECIMAL(9,2),
CDECIMAL10 	DECIMAL(10),
CDECIMAL18 	DECIMAL(18),
CDECIMAL18_2 	DECIMAL(18,2),
CINTEGER 	INTEGER,
CSMALLINT 	SMALLINT,
CTINYINT 	TINYINT,
CDOUBLE 	DOUBLE,
CFLOAT		FLOAT,
CDATE 		DATE,
CDATETIME 	DATETIME,
CTIME	TIME,
CCHAR1 	CHAR(1),
CCHAR2 	CHAR(2),
CCHAR3 	CHAR(3),
CCHAR4 	CHAR(4),
CCHAR5 	CHAR(5),
CCHAR6 	CHAR(6),
CCHAR7 	CHAR(7),
CCHAR8 	CHAR(8),
CCHAR9 	CHAR(9),
CCHAR255 	CHAR(255),
CVCHAR1 	VARCHAR(1),
CVCHAR2 	VARCHAR(2),
CVCHAR3 	VARCHAR(3),
CVCHAR4 	VARCHAR(4),
CVCHAR5 	VARCHAR(5),
CVCHAR6 	VARCHAR(6),
CVCHAR7 	VARCHAR(7),
CVCHAR8 	VARCHAR(8),
CVCHAR255 	VARCHAR(255),
CTEXT TEXT,
CBLOB BLOB,
CMEDIUMINT MEDIUMINT,
CTINYTEXT TINYTEXT,
CTINYBLOB TINYBLOB,
CMEDIUMTEXT MEDIUMTEXT,
CMEDIUMBLOB MEDIUMBLOB,
CLONGTEXT LONGTEXT,
CLONGBLOB LONGBLOB,
CTIMESTAMP TIMESTAMP
) engine=Columnstore;
#
insert into datatypeupdate values (1,-72036854775806,-9,-999,-9.99,-999,-999999,-9999.99,-9999999,-9999999999999,-99999999999.99,-7483646,-766,-26,-1.797693231E+108,-3.402866E+18,'1997-01-01','1997-01-01 00:00:00','13:00:00','a','aa','aaa','aaaa','aaaaa','aaaaaa','aaaaaaa','aaaaaaaa','aaaaaaaaa','aaaaaaaaaa','a','aa','aaa','aaaa','aaaaa','aaaaaa','aaaaaaa','aaaaaaaa','aaaaaaaaaa','texttexttext','blobblobblob',1000,'tinytext','tinyblob','mediumtestmediumtest','mediumblobmediumblob','longtextlongtextlongtext','longbloblongbloblongblob','2020-08-13 03:14:07');
