/* Copyright (C) MariaDB Corporation Ab */
#define MSG_ACCESS_VIOLATN  "Access violation"
#define MSG_ADD_BAD_TYPE    "Array add value type mismatch (%s -> %s)"
#define MSG_ALLOC_ERROR     "Error allocating %s"
#define MSG_ANSWER_TYPE     "Answer of type"
#define MSG_API_CONF_ERROR  "SQL Error: API_CONFORMANCE"
#define MSG_APPL_NOT_INIT   "Application not initialized"
#define MSG_ARRAY_BNDS_EXCD "Array bounds exceeded"
#define MSG_BAD_ARRAY_OPER  "Arrays must be used with the IN operator"
#define MSG_BAD_ARRAY_TYPE  "Illegal array type %d"
#define MSG_BAD_ARRAY_VAL   "Arrays must have the same number of values"
#define MSG_BAD_BIN_FMT     "Invalid format %c for the %s BIN column"
#define MSG_BAD_BLK_ESTIM   "Number of blocks exceeds estimate"
#define MSG_BAD_BLK_SIZE    "No match in block %d size"
#define MSG_BAD_BYTE_NUM    "bad number of bytes written"
#define MSG_BAD_BYTE_READ   "bad number of bytes read"
#define MSG_BAD_COL_TYPE    "Invalid type %s for column %s"
#define MSG_BAD_COL_XPATH   "Invalid Xpath in column %s for HTML table %s"
#define MSG_BAD_CONST_TYPE  "Bad constant type=%d"
#define MSG_BAD_CONV_TYPE   "Invalid convert type %d"
#define MSG_BAD_DATETIME    "Invalid datetime value"
#define MSG_BAD_DBF_FILE    "DBF file %s is corrupted"
#define MSG_BAD_DBF_REC     "DBF file %s corrupted at record %d"
#define MSG_BAD_DBF_TYPE    "Unsupported DBF type %c for column %s"
#define MSG_BAD_DIRECTORY   "Bad directory %s: %s"
#define MSG_BAD_FIELD_RANK  "Invalid field rank %d for column %s"
#define MSG_BAD_FIELD_TYPE  "Bad type field %s"
#define MSG_BAD_FILE_HANDLE "Invalid File Handle: %s"
#define MSG_BAD_FILTER      "Bad filter: Opc=%d B_T=%d %d Type=%d %d"
#define MSG_BAD_FILTER_CONV "Bad filter conversion, B_T=%d,%d"
#define MSG_BAD_FILTER_OP   "Invalid filter operator %d"
#define MSG_BAD_FLD_FORMAT  "Bad format for field %d of %s"
#define MSG_BAD_FLD_LENGTH  "Field %s too long (%s --> %d) line %d of %s"
#define MSG_BAD_FREQ_SET    "Bad frequency setting for column %s"
#define MSG_BAD_FUNC_MODE   "%s: invalid mode %d"
#define MSG_BAD_HANDLE_VAL  "Invalid handle value"
#define MSG_BAD_HEADER      "File %s: Header corrupted"
#define MSG_BAD_HEAD_END    "Can't read end of header"
#define MSG_BAD_INDEX_FILE  "Wrong index file %s"
#define MSG_BAD_LINEFLD_FMT "Bad format line %d field %d of %s"
#define MSG_BAD_LINE_LEN    "Line length not equal to Lrecl"
#define MSG_BAD_LRECL       "Table/File lrecl mismatch (%d,%hd)"
#define MSG_BAD_NODE_TYPE   "Bad type %d for table node"
#define MSG_BAD_OFFSET_VAL  "Invalid null offset value for a CSV table"
#define MSG_BAD_OPEN_MODE   "Invalid open mode %d"
#define MSG_BAD_PARAM_TYPE  "%.8s: Bad parameter type=%d"
#define MSG_BAD_PARM_COUNT  "Parameter count mismatch"
#define MSG_BAD_QUOTE_FIELD "Missing ending quote in %s field %d line %d"
#define MSG_BAD_READ_NUMBER "Wrong number %d of values read from %s"
#define MSG_BAD_RECFM       "Invalid recfm type %d for DOSCOL"
#define MSG_BAD_RECFM_VAL   "Bad Recfm value %d"
#define MSG_BAD_SET_CASE    "Cannot set sensitive an insensitive array"
#define MSG_BAD_SET_STRING  "Invalid SetValue from string"
#define MSG_BAD_SPECIAL_COL "Bad special column %s"
#define MSG_BAD_SPEC_COLUMN "Special column invalid for this table type"
#define MSG_BAD_TABLE_TYPE  "Bad type %s for table %s"
#define MSG_BAD_TYPE_LIKE   "Bad operand(%d) type=%d for LIKE"
#define MSG_BAD_VALBLK_INDX "Out of range valblock index value"
#define MSG_BAD_VALBLK_TYPE "Invalid value block type %d"
#define MSG_BAD_VALNODE     "Bad type %d for column %s value node"
#define MSG_BAD_VALUE_TYPE  "Invalid value type %d"
#define MSG_BAD_VAL_UPDATE  "Don't know which %s value to update"
#define MSG_BAS_NS_LIST     "Invalid namespaces list format"
#define MSG_BIN_F_TOO_LONG  "Value too long for field %s (%d --> %d)"
#define MSG_BIN_MODE_FAIL   "Set binary mode failed: %s"
#define MSG_BLKTYPLEN_MISM  "Non matching block types/lengths in SetValue"
#define MSG_BLK_IS_NULL     "Blk is NULL"
#define MSG_BREAKPOINT      "Breakpoint"
#define MSG_BUILD_INDEX     "Building index %s on %s"
#define MSG_CANNOT_OPEN     "Cannot open %s"
#define MSG_CHSIZE_ERROR    "chsize error: %s"
#define MSG_COL_ALLOC_ERR   "Cannot allocate column node"
#define MSG_COL_ISNOT_TABLE "Column %s is not in table %s"
#define MSG_COL_NOT_SORTED  "Column %s of table %s is not sorted"
#define MSG_COL_NUM_MISM    "Number of columns mismatch"
#define MSG_COM_ERROR       "Com error"
#define MSG_CONCAT_SUBNODE  "Cannot concatenate sub-nodes"
#define MSG_CONNECT_CANCEL  "Connection cancelled by user"
#define MSG_CONTROL_C_EXIT  "Control C exit"
#define MSG_DATABASE_LOADED "Database %s loaded"
#define MSG_DATA_MISALIGN   "Datatype misalignment"
#define MSG_DBASE_FILE      "dBASE dbf file: "
#define MSG_DEF_ALLOC_ERROR "Error allocating %s DEF class"
#define MSG_DEL_FILE_ERR    "Error deleting %s"
#define MSG_DEL_READ_ERROR  "Delete: read error req=%d len=%d"
#define MSG_DEL_WRITE_ERROR "Delete: write error: %s"
#define MSG_DEPREC_FLAG     "Deprecated option Flag, use Coltype"
#define MSG_DLL_LOAD_ERROR  "Error %d loading module %s"
#define MSG_DOM_NOT_SUPP    "MS-DOM not supported by this version"
#define MSG_DVAL_NOTIN_LIST "Value %s not found in distinct values list of column %s"
#define MSG_EMPTY_DOC       "Empty document"
#define MSG_EMPTY_FILE      "%s empty file %s: "
#define MSG_EOF_AFTER_LINE  "EOF after line %d"
#define MSG_EOF_INDEX_FILE  "EOF while reading index file"
#define MSG_ERROR_IN_LSK    "Error %d in lseek64"
#define MSG_ERROR_IN_SFP    "Error %d in SetFilePointer"
#define MSG_ERR_READING_REC "Error reading record %d of %s"
#define MSG_FAIL_ADD_NODE   "Failed to add %s table node"
#define MSG_FETCH_NO_RES    "Fetch: No Result Set"
#define MSG_FIELD_TOO_LONG  "Value too long for field %d line %d"
#define MSG_FILELEN_ERROR   "Error in %s for %s"
#define MSG_FILE_IS_EMPTY   "File %s is empty"
#define MSG_FILE_MAP_ERR    "File mapping error"
#define MSG_FILE_MAP_ERROR  "CreateFileMapping %s error rc=%d"
#define MSG_FILE_OPEN_YET   "File %s already open"
#define MSG_FILE_UNFOUND    "File %s not found"
#define MSG_FIX_OVFLW_ADD   "Fixed Overflow on add"
#define MSG_FIX_OVFLW_TIMES "Fixed Overflow on times"
#define MSG_FIX_UNFLW_ADD   "Fixed Underflow on add"
#define MSG_FIX_UNFLW_TIMES "Fixed Underflow on times"
#define MSG_FLD_TOO_LNG_FOR "Field %d too long for %s line %d of %s"
#define MSG_FLT_BAD_RESULT  "Float inexact result"
#define MSG_FLT_DENORMAL_OP "Float denormal operand"
#define MSG_FLT_INVALID_OP  "Float invalid operation"
#define MSG_FLT_OVERFLOW    "Float overflow"
#define MSG_FLT_STACK_CHECK "Float stack check"
#define MSG_FLT_UNDERFLOW   "Float underflow"
#define MSG_FLT_ZERO_DIVIDE "Float divide by zero"
#define MSG_FMT_WRITE_NIY   "Writing %s files is not implemented yet"
#define MSG_FOXPRO_FILE     "FoxPro file: "
#define MSG_FPUTS_ERROR     "fputs error: %s"
#define MSG_FSEEK_ERROR     "fseek error: %s"
#define MSG_FSETPOS_ERROR   "fseek error for i=%d"
#define MSG_FTELL_ERROR     "ftell error for recd=%d: %s"
#define MSG_FUNCTION_ERROR  "%s error: %d"
#define MSG_FUNC_ERRNO      "Error %d in %s"
#define MSG_FUNC_ERROR      "Error in %s"
#define MSG_FUNC_ERR_S      "%s error: %s"
#define MSG_FWRITE_ERROR    "fwrite error: %s"
#define MSG_GET_DIST_VALS   "Retrieving distinct values from "
#define MSG_GET_FUNC_ERR    "Error getting function %s: %s"
#define MSG_GLOBAL_ERROR    "Cannot allocate Global (size=%d)\n"
#define MSG_GUARD_PAGE      "Guard page violation"
#define MSG_GZOPEN_ERROR    "gzopen %s error %d on %s"
#define MSG_ILLEGAL_INSTR   "Illegal instruction"
#define MSG_ILL_FILTER_CONV "Filtering implies an illegal conversion"
#define MSG_INDEX_NOT_UNIQ  "Index is not unique"
#define MSG_INDEX_YET_ON    "Index %s already exists on %s"
#define MSG_INDX_COL_NOTIN  "Index column %s is not in table %s"
#define MSG_INDX_EXIST_YET  "Index entry already exists"
#define MSG_INIT_FAILED     "Failed to initialize %s processing"
#define MSG_INT_COL_ERROR   "Internal error for index column %s"
#define MSG_INT_OVERFLOW    "Integer overflow"
#define MSG_INT_ZERO_DIVIDE "Integer divide by zero"
#define MSG_INVALID_DISP    "Invalid disposition"
#define MSG_INVALID_FTYPE   "SBV: invalid Ftype %d"
#define MSG_INVALID_HANDLE  "Invalid handle"
#define MSG_INVALID_OPER    "Invalid operator %d for %s"
#define MSG_INV_COLUMN_TYPE "Invalid type %d for column %s"
#define MSG_INV_COL_TYPE    "Invalid column type %s"
#define MSG_INV_DEF_READ    "Invalid deferred Read rc=%d"
#define MSG_INV_DIRCOL_OFST "Invalid DIRCOL offset %d"
#define MSG_INV_MAP_POS     "Invalid map position"
#define MSG_INV_RAND_ACC    "Invalid random access to non optimized table"
#define MSG_INV_REC_POS     "Invalid record position"
#define MSG_INV_RESULT_TYPE "Invalid result type %s"
#define MSG_INV_UPDT_TABLE  "Table %s invalid for update"
#define MSG_IN_WITHOUT_SUB  "IN or EXISTS without array or subquery"
#define MSG_KEY_ALLOC_ERR   "Error allocating Key offset block"
#define MSG_KEY_ALLOC_ERROR "Memory allocation error, Klen=%d n=%d"
#define MSG_LINE_TOO_LONG   "New line is too long"
#define MSG_LIST            "--List--"
#define MSG_LOADING_FAILED  "Loading of %s failed"
#define MSG_LRECL_TOO_SMALL "Lrecl too small (headlen = %d)"
#define MSG_MAKE_EMPTY_FILE "Making empty file %s: %s"
#define MSG_MAKING          "Making"
#define MSG_MALLOC_ERROR    "Memory allocation failed: %s returned Null"
#define MSG_MAP_VIEW_ERROR  "MapViewOfFile %s error rc=%d"
#define MSG_MAXSIZE_ERROR   "Cannot calculate max size on open table"
#define MSG_MEM_ALLOC_ERR   "Memory allocation error, %s size=%d"
#define MSG_MEM_ALLOC_ERROR "Memory allocation error"
#define MSG_MISPLACED_QUOTE "Misplaced quote in line %d"
#define MSG_MISSING_ARG     "Missing argument for operator %d"
#define MSG_MISSING_FIELD   "Missing field %d in %s line %d"
#define MSG_MISSING_FNAME   "Missing file name"
#define MSG_MISSING_NODE    "Missing %s node in %s"
#define MSG_MISSING_ROWNODE "Can't find RowNode for row %d"
#define MSG_MIS_TAG_LIST    "Missing column tag list"
#define MSG_MUL_MAKECOL_ERR "Tabmul MakeCol logical error"
#define MSG_NAME_CONV_ERR   "Error converting node name"
#define MSG_NEW_DOC_FAILED  "Cannot create new document"
#define MSG_NEW_RETURN_NULL "New returned Null in PlugEvalLike"
#define MSG_NEXT_FILE_ERROR "Couldn't find next file. rc=%d"
#define MSG_NONCONT_EXCEPT  "Noncontinuable exception"
#define MSG_NOP_ZLIB_INDEX  "Cannot do indexing on non optimized zlib table"
#define MSG_NOT_A_DBF_FILE  "Not a dBASE dbf file "
#define MSG_NOT_FIXED_LEN   "File %s is not fixed length, len=%d lrecl=%d"
#define MSG_NO_0DH_HEAD     "No 0Dh at end of header (dbc=%d)"
#define MSG_NO_ACTIVE_DB    "No active database"
#define MSG_NO_CHAR_FROM    "Cannot return char value from type %d"
#define MSG_NO_DATE_FMT     "No date format for valblock of type %d"
#define MSG_NO_DEF_FNCCOL   "Cannot find default function column"
#define MSG_NO_DEF_PIVOTCOL "Cannot find default pivot column"
#define MSG_NO_DIR_INDX_RD  "No direct access of %s tables"
#define MSG_NO_FEAT_SUPPORT "No %s support in this version"
#define MSG_NO_FLD_FORMAT   "Missing format for field %d of %s"
#define MSG_NO_FORMAT_COL   "Cannot format the type COLUMN"
#define MSG_NO_FORMAT_TYPE  "Cannot set format from type %d"
#define MSG_NO_INDEX_READ   "No indexed read for multiple tables"
#define MSG_NO_KEY_COL      "No key columns found"
#define MSG_NO_KEY_UPDATE   "Cannot update key names"
#define MSG_NO_MAP_INSERT   "MAP incompatible with Insert"
#define MSG_NO_MATCHING_COL "No matching column %s in %s"
#define MSG_NO_MATCH_COL    "Cannot find matching column"
#define MSG_NO_MEMORY       "No memory"
#define MSG_NO_MODE_PADDED  "Mode not supported for padded files"
#define MSG_NO_MUL_VCT      "VCT tables cannot be multiple"
#define MSG_NO_ODBC_DELETE  "Delete should not be called for ODBC tables"
#define MSG_NO_ODBC_DIRECT  "Direct access of ODBC tables not implemented yet"
#define MSG_NO_ODBC_MUL     "Multiple(2) not supported for ODBC tables"
#define MSG_NO_ODBC_SPECOL  "No ODBC special columns"
#define MSG_NO_PART_DEL     "No partial delete of %s files"
#define MSG_NO_PART_MAP     "Partial mapping not implemented for this OS"
#define MSG_NO_PAR_BLK_INS  "Cannot insert partial block yet"
#define MSG_NO_PIV_DIR_ACC  "No direct access to PIVOT tables"
#define MSG_NO_READ_32      "Can't read 32 bytes"
#define MSG_NO_RECOV_SPACE  "Cannot recover space in index file"
#define MSG_NO_ROWID_FOR_AM "Can't get RowID in direct access for tables of type %s"
#define MSG_NO_ROW_NODE     "Row node name is not defined"
#define MSG_NO_SECTION_NAME "Missing section name"
#define MSG_NO_SEC_UPDATE   "Cannot update section names"
#define MSG_NO_SETPOS_YET   "%s SetPos not implemented yet"
#define MSG_NO_SPEC_COL     "No MySQL special columns"
#define MSG_NO_SUB_VAL      "No sub value for array of type %d"
#define MSG_NO_TABCOL_DATA  "No data found for table %s column %s"
#define MSG_NO_TABLE_DEL    "Delete not enabled for %s tables   "
#define MSG_NO_TAB_DATA     "No data found for table %s"
#define MSG_NO_VCT_DELETE   "Partial delete not yet implemented for VCT files"
#define MSG_NO_ZIP_DELETE   "Delete Zip files not implemented yet"
#define MSG_OPENING         "Opening"
#define MSG_OPEN_EMPTY_FILE "Opening empty file %s: %s"
#define MSG_OPEN_ERROR      "Open error %d in mode %d on %s: "
#define MSG_OPEN_ERROR_IS   "Open error on %s: %s"
#define MSG_OPEN_MODE_ERROR "Open(%s) error %d on %s"
#define MSG_OPEN_STRERROR   "open error: %s"
#define MSG_OPTBLK_RD_ERR   "Error reading opt block values: %s"
#define MSG_OPTBLK_WR_ERR   "Error writing opt block values: %s"
#define MSG_OPTIMIZING      "Optimizing "
#define MSG_OPT_BMAP_RD_ERR "Error reading opt bitmaps: %s"
#define MSG_OPT_BMAP_WR_ERR "Error writing opt bitmaps: %s"
#define MSG_OPT_CANCELLED   "Optimize cancelled by User"
#define MSG_OPT_DVAL_RD_ERR "Error reading distinct values: %s"
#define MSG_OPT_DVAL_WR_ERR "Error writing distinct values: %s"
#define MSG_OPT_HEAD_RD_ERR "Error reading opt file header: %s"
#define MSG_OPT_HEAD_WR_ERR "Error writing opt file header: %s"
#define MSG_OPT_LOGIC_ERR   "Logical error in SetBitmap, i=%d"
#define MSG_OPT_MAX_RD_ERR  "Error reading opt max values: %s"
#define MSG_OPT_MAX_WR_ERR  "Error writing opt max values: %s"
#define MSG_OPT_MIN_RD_ERR  "Error reading opt min values: %s"
#define MSG_OPT_MIN_WR_ERR  "Error writing opt min values: %s"
#define MSG_OPT_NOT_MATCH   "Non-matching opt file %s"
#define MSG_PAGE_ERROR      "In page error"
#define MSG_PARM_CNT_MISS   "Parameter count mismatch"
#define MSG_PREC_VBLP_NULL  "ARRAY SetPrecision: Vblp is NULL"
#define MSG_PRIV_INSTR      "Privileged instruction"
#define MSG_PROCADD_ERROR   "Error %d getting address of %s"
#define MSG_QUERY_CANCELLED "Query Cancelled by User"
#define MSG_RANGE_NO_JOIN   "Range is not meant for join index"
#define MSG_RC_READING      "rc=%d reading table %s"
#define MSG_READY           "Ready"
#define MSG_READ_ERROR      "Error reading %s: %s"
#define MSG_READ_ONLY       "Cannot modify this read/only protected table"
#define MSG_READ_SEEK_ERROR "Read seek error: %s"
#define MSG_REGISTER_ERR    "Unable to register NS with prefix='%s' and href='%s'"
#define MSG_REMOVE_ERROR    "Error removing %s: %s"
#define MSG_RENAME_ERROR    "Error renaming %s to %s: %s"
#define MSG_ROWID_NOT_IMPL  "RowNumber not implemented for tables of type %s"
#define MSG_SEC_KEY_FIRST   "Section and key names must come first on Insert"
#define MSG_SEC_NAME_FIRST  "Section name must come first on Insert"
#define MSG_SEP_IN_FIELD    "Field %d contains the separator character"
#define MSG_SEQUENCE_ERROR  "Sequence error on statement allocation"
#define MSG_SETEOF_ERROR    "Error %d in SetEndOfFile"
#define MSG_SETRECPOS_NIY   "SetRecpos not implemented for this table type"
#define MSG_SET_STR_TRUNC   "SetValue: String would be truncated"
#define MSG_SFP_ERROR       "SetFilePointer error: %s"
#define MSG_SHARED_LIB_ERR  "Error loading shared library %s: %s"
#define MSG_SINGLE_STEP     "Single step"
#define MSG_SORTING_VAL     "Sorting %d values"
#define MSG_SPCOL_READONLY  "Special column %s is Read Only"
#define MSG_SQL_CONF_ERROR  "SQL Error: SQL_CONFORMANCE"
#define MSG_SRCH_CLOSE_ERR  "Couldn't close search handle"
#define MSG_SRC_TABLE_UNDEF "Source table is not defined"
#define MSG_STACK_OVERFLOW  "Stack overflow"
#define MSG_TABDIR_READONLY "DIR tables are read/only"
#define MSG_TABLE_NOT_OPT   "Not an optimizable table"
#define MSG_TABLE_NO_INDEX  "Table %s is not indexable"
#define MSG_TABLE_READ_ONLY "%s tables are read only   "
#define MSG_TABMUL_READONLY "Multiple tables are read/only"
#define MSG_TOO_MANY_FIELDS "Too many fields line %d of %s"
#define MSG_TOO_MANY_JUMPS  "Too many jump levels"
#define MSG_TOO_MANY_KEYS   "Too many keys (%d)"
#define MSG_TO_BLK_IS_NULL  "To Blk is NULL"
#define MSG_TRUNCATE_ERROR  "truncate error: %s"
#define MSG_TRUNC_BY_ESTIM  "truncated by Estimate"
#define MSG_TYPE_MISMATCH   "Key and source are not of the same type"
#define MSG_TYPE_VALUE_ERR  "Column %s type(%s)/value(%s) mismatch"
#define MSG_UNBALANCE_QUOTE "Unbalanced quote in line %d"
#define MSG_UNDEFINED_AM    "COLBLK %s: undefined Access Method"
#define MSG_UNKNOWN_EXCPT   "Unknown exception"
#define MSG_UNMATCH_FIL_ARG "Unmatched filter argument"
#define MSG_UPDATE_ERROR    "Error updating %s"
#define MSG_UPD_ZIP_NOT_IMP "Updating ZDOS tables not implemented yet"
#define MSG_VALSTR_TOO_LONG "Value %s too long for string of length %d"
#define MSG_VALTYPE_NOMATCH "Non matching Value types"
#define MSG_VALUE_ERROR     "Column %s: value is null"
#define MSG_VALUE_TOO_BIG   "Value %lld too big for column %s"
#define MSG_VALUE_TOO_LONG  "Value %s too long for column %s of length %d"
#define MSG_VAL_ALLOC_ERR   "Cannot allocate value node"
#define MSG_VIR_NO_DELETE   "Delete not allowed for %s tables"
#define MSG_VIR_READ_ONLY   "Virtual %s tables are read only"
#define MSG_VOID_FIRST_ARG  "First argument should not be void"
#define MSG_WORK_AREA       "Work area: %s"
#define MSG_WRITE_SEEK_ERR  "Write seek error: %s"
#define MSG_WRITE_STRERROR  "Error writing %s: %s"
#define MSG_WRITING         "Writing"
#define MSG_WRITING_ERROR   "Error writing to %s: %s"
#define MSG_WS_CONV_ERR     "Error converting %s to WS"
#define MSG_XCOL_MISMATCH   "Column %s mismatch in index"
#define MSG_XFILE_READERR   "Error %d reading index file"
#define MSG_XFILE_WRITERR   "Error writing index file: %s"
#define MSG_XMLTAB_INIT_ERR "Error initializing XML table"
#define MSG_XML_INIT_ERROR  "Error initializing new XML file"
#define MSG_XPATH_CNTX_ERR  "Unable to create new XPath context"
#define MSG_XPATH_EVAL_ERR  "Unable to evaluate xpath location '%s'"
#define MSG_XPATH_NOT_SUPP  "Unsupported Xpath for column %s"
#define MSG_ZERO_DIVIDE     "Zero divide in expression"
#define MSG_ARG_IS_NULL     "Argument is NULL"
