#define IDS_TABLES             100
#define IDS_TAB_01             101
#define IDS_TAB_02             102
#define IDS_TAB_03             103
#define IDS_TAB_04             104
#define IDS_TAB_05             105
#define IDS_COLUMNS            106
#define IDS_COL_01             107
#define IDS_COL_02             108
#define IDS_COL_03             109
#define IDS_COL_04             110
#define IDS_COL_05             111
#define IDS_COL_06             112
#define IDS_COL_07             113
#define IDS_COL_08             114
#define IDS_COL_09             115
#define IDS_COL_10             116
#define IDS_COL_11             117
#define IDS_COL_12             118
#define IDS_PKEY               119
#define IDS_PKY_01             120
#define IDS_PKY_02             121
#define IDS_PKY_03             122
#define IDS_PKY_04             123
#define IDS_PKY_05             124
#define IDS_PKY_06             125
#define IDS_STAT               126
#define IDS_STA_01             127
#define IDS_STA_02             128
#define IDS_STA_03             129
#define IDS_STA_04             130
#define IDS_STA_05             131
#define IDS_STA_06             132
#define IDS_STA_07             133
#define IDS_STA_08             134
#define IDS_STA_09             135
#define IDS_STA_10             136
#define IDS_STA_11             137
#define IDS_STA_12             138
#define IDS_STA_13             139
#define IDS_DRIVER             140
#define IDS_DRV_01             141
#define IDS_DRV_02             142
#define IDS_DSRC               143
#define IDS_DSC_01             144
#define IDS_DSC_02             145
