CREATE TABLE `supplier` (
  `S_SUPPKEY` int(11) NOT NULL,
  `S_NAME` char(25) NOT NULL,
  `S_ADDRESS` varchar(40) NOT NULL,
  `S_NATIONKEY` int(11) NOT NULL,
  `S_PHONE` char(15) NOT NULL,
  `S_ACCTBAL` decimal(15,2) NOT NULL,
  `S_COMMENT` varchar(101) NOT NULL,
  KEY `S_NATIONKEY` (`S_NATIONKEY`,`S_SUPPKEY`,`S_NAME`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb3 COLLATE=utf8mb3_general_ci;
INSERT INTO `supplier` (
    `S_SUPPKEY`, `S_NAME`, `S_ADDRESS`, `S_NATIONKEY`, `S_PHONE`, `S_ACCTBAL`, `S_COMMENT`
) VALUES
(1855, 'Supplier#000001855', 'MWk6EAeozXb', 3, '13-710-803-7566', 46.35, ' sleep furiously. unusual instruc'),
(2204, 'Supplier#000002204', 'uYmlr46C06udCqanj0KiRsoTQakZsEyssL', 3, '13-529-779-6608', 580.84, 'uthlessly express, regular deposits. slyly special asymptotes must have to are'),
(2218, 'Supplier#000002218', 'nODZw5q4dx kp0K5', 3, '13-587-588-9916', 2164.76, ' fluffily even requests sleep quickly'),
(2282, 'Supplier#000002282', 'ES21K9dxoW1I1TzWCj7ekdlNwSWnv1Z  6mQ,BKn', 3, '13-882-982-2553', 9689.51, 's sleep blithely after the fluffily ironic asymptotes. quickly pending instructions slee'),
(2419, 'Supplier#000002419', 'qydBQd14I5l5mVXa4fYY', 3, '13-362-261-8296', 2080.92, 'alongside of the dependencies. theodolites affix deposits. boldly regular packages sleep against t'),
(2799, 'Supplier#000002799', 'lwr, 6L3gdfc79PQut,4XO6nQsTJY63cAyYO', 3, '13-757-175-8800', 8439.97, 'en, quick ideas detect quietly against'),
(2941, 'Supplier#000002941', 'Naddba 8YTEKekZyP0', 3, '13-722-750-8709', 8001.44, 'l asymptotes. express plate'),
(3185, 'Supplier#000003185', 'hMa535Cbf2mj1Nw4OWOKWVrsK0VdDkJURrdjSIJe', 3, '13-979-274-8468', -475.90, 'quickly final pinto beans wake carefu'),
(3422, 'Supplier#000003422', 'DJoCEapUeBXoV1iYiCcPFQvzsTv2ZI960', 3, '13-662-152-4869', 5731.96, 'xes sleep blithely ironic requests. slyly express deposits boost furiou'),
(5300, 'Supplier#000005300', 'gXG28YqpxU', 3, '13-790-538-6012', 4454.86, 'y even theodolites boost. final foxes unwind. regular packages haggle carefully idea'),
(5426, 'Supplier#000005426', '9Dz2OVT1q sb4BK71ljQ1XjPBYRPvO', 3, '13-448-787-4918', 7712.02, ' fluffily slyly regular packages. blithely final pl'),
(6093, 'Supplier#000006093', 'KJNUg1odUT2wtCS2s6PrH3D6fd', 3, '13-989-796-2627', 7935.95, 'ts solve slyly slyly final requests. slyly regular accounts haggle regul'),
(6109, 'Supplier#000006109', 'rY5gbfh3dKHnylcQUTPGCwnbe', 3, '13-232-509-4334', 5343.69, 'the slyly unusual platelets. always ironic accounts mold blithely. requests use slyly final'),
(6949, 'Supplier#000006949', 'mLxYUJhsGcLtKe ,GFirNu183AvT', 3, '13-776-634-2532', 2792.37, 'boldly bold gifts are deposits. even accou'),
(7169, 'Supplier#000007169', 'tEc95D2moN9S84nd55O,dlnW', 3, '13-805-131-6933', 9299.64, ' slyly against the pending platelets. final, final pa'),
(7801, 'Supplier#000007801', '69fi,U1r6enUb  ', 3, '13-235-541-2569', 1296.61, 'lar, even deposits boost. final accounts serve after the furiously pen'),
(8595, 'Supplier#000008595', 'MH0iB73GQ3z UW3O DbCbqmc', 3, '13-841-273-9262', 2195.24, 'ily silent deposits into the slyly silent request'),
(8610, 'Supplier#000008610', 'SgVgP90vP452sUNTgzL9zKwXHXAzV6tV', 3, '13-107-134-7845', 307.37, 'lar theodolites sleep carefully'),
(9430, 'Supplier#000009430', 'igRqmneFt  ', 3, '13-491-379-4658', 3345.48, 'aggle blithely at the silent warthogs. regular foxes cajole along th'),
(9899, 'Supplier#000009899', '7XdpAHrzr1t,UQFZE', 3, '13-200-446-4635', 4235.99, 'print blithely. carefully final theodolites alongside of the blithely even'),
(530, 'Supplier#000000530', '0qwCMwobKY OcmLyfRXlagA8ukENJv,', 3, '13-747-781-9694', 4327.86, 'fily pending dependencies wake slyly final deposits. platelets against the slyly ironic requests na'),
(729, 'Supplier#000000729', 'pqck2ppy758TQpZCUAjPvlU55K3QjfL7Bi', 3, '13-627-404-3241', 7113.46, 'haggle furiously among the express accounts. ironic, express warhorses promise. even, expr'),
(1245, 'Supplier#000001245', 'bztTCfHZnsLafgtZWe,isV Gr 4IFg', 3, '13-597-902-8280', 9851.62, 's. accounts haggle slyly. f'),
(1700, 'Supplier#000001700', '7hMlCof1Y5zLFg', 3, '13-910-616-9417', 5786.59, 'ly pending pinto beans are-- final asymptotes haggl'),
(1730, 'Supplier#000001730', 'Rc8e,1Pybn r6zo0VJIEiD0UD vhk', 3, '13-788-703-6101', 8712.72, ' pending ideas are carefully furiously unusual asymptotes. express packag'),
(2204, 'Supplier#000002204', 'uYmlr46C06udCqanj0KiRsoTQakZsEyssL', 3, '13-529-779-6608', 580.84, 'uthlessly express, regular deposits. slyly special asymptotes must have to are'),
(2218, 'Supplier#000002218', 'nODZw5q4dx kp0K5', 3, '13-587-588-9916', 2164.76, ' fluffily even requests sleep quickly'),
(2282, 'Supplier#000002282', 'ES21K9dxoW1I1TzWCj7ekdlNwSWnv1Z  6mQ,BKn', 3, '13-882-982-2553', 9689.51, 's sleep blithely after the fluffily ironic asymptotes. quickly pending instructions slee'),
(2303, 'Supplier#000002303', 'nCoWfpB6YOymbgOht7ltfklpkHl', 3, '13-358-165-6810', 1984.35, 'silent theodolites boost quickly furiously even pinto beans. blithely ironic orbits promise'),
(2419, 'Supplier#000002419', 'qydBQd14I5l5mVXa4fYY', 3, '13-362-261-8296', 2080.92, 'alongside of the dependencies. theodolites affix deposits. boldly regular packages sleep against t'),
(2629, 'Supplier#000002629', '0Bw,q5Zp8su9XrzoCngZ3cAEXZwZ', 3, '13-192-161-4976', 3633.93, 's. even, express excuses after the deposits are carefu'),
(2799, 'Supplier#000002799', 'lwr, 6L3gdfc79PQut,4XO6nQsTJY63cAyYO', 3, '13-757-175-8800', 8439.97, 'en, quick ideas detect quietly against'),
(2941, 'Supplier#000002941', 'Naddba 8YTEKekZyP0', 3, '13-722-750-8709', 8001.44, 'l asymptotes. express plate'),
(3185, 'Supplier#000003185', 'hMa535Cbf2mj1Nw4OWOKWVrsK0VdDkJURrdjSIJe', 3, '13-979-274-8468', -475.90, 'quickly final pinto beans wake carefu'),
(3201, 'Supplier#000003201', 'E87yws6I,t0qNs4QW7UzExKiJnJDZWue', 3, '13-600-413-7165', 3799.41, 'ely final asymptotes. express realms sleep. packages cajole; carefully sil'),
(3288, 'Supplier#000003288', 'EDdfNt7E5Uc,xLTupoIgYL4yY7ujh,', 3, '13-442-617-3554', -79.72, 'nt deposits. special courts serve f'),
(3373, 'Supplier#000003373', 'iy8VM48ynpc3N2OsBwAvhYakO2us9R1bi', 3, '13-951-330-3925', 9457.95, 'furiously final pinto beans detect about the pinto beans. ironic pinto beans wake among the re'),
(3380, 'Supplier#000003380', 'jPv0V,pszouuFT3YsAqlP,kxT3u,gTFiEbRt,x', 3, '13-827-850-8102', 1340.88, 'nts of the unusual foxes cajole furiously packages. foxes detec'),
(3422, 'Supplier#000003422', 'DJoCEapUeBXoV1iYiCcPFQvzsTv2ZI960', 3, '13-662-152-4869', 5731.96, 'xes sleep blithely ironic requests. slyly express deposits boost furiou'),
(3528, 'Supplier#000003528', 'qJB,nuoxi0AGuX,9Z3tzDK', 3, '13-249-999-5949', 9486.63, 'ress fluffily. quickly special platelets sleep. ironic, final wa'),
(4014, 'Supplier#000004014', 'Spla6 0RV3oK9Y', 3, '13-587-428-2644', 772.46, ' deposits? slyly bold accounts wake carefully above the carefully special packages. f'),
(4201, 'Supplier#000004201', 'C5x43M76dnb1w2LFbhB', 3, '13-431-798-5549', 6735.20, 'ies use furiously. carefully unusual theodolites boost slyly'),
(4304, 'Supplier#000004304', 'hQCAz59k,HLlp2CKUrcBIL', 3, '13-419-753-1541', 630.97, 'ly unusual frets? quickly even sentiments nag. quickly final deposit'),
(4522, 'Supplier#000004522', 'xXtCKwsZDArxIBGDfzX2PgobGZsBg', 3, '13-691-847-9748', 7453.60, 'eas. always blithe deposits detect slyly after the'),
(5300, 'Supplier#000005300', 'gXG28YqpxU', 3, '13-790-538-6012', 4454.86, 'y even theodolites boost. final foxes unwind. regular packages haggle carefully idea'),
(5426, 'Supplier#000005426', '9Dz2OVT1q sb4BK71ljQ1XjPBYRPvO', 3, '13-448-787-4918', 7712.02, ' fluffily slyly regular packages. blithely final pl'),
(6093, 'Supplier#000006093', 'KJNUg1odUT2wtCS2s6PrH3D6fd', 3, '13-989-796-2627', 7935.95, 'ts solve slyly slyly final requests. slyly regular accounts haggle regul'),
(6109, 'Supplier#000006109', 'rY5gbfh3dKHnylcQUTPGCwnbe', 3, '13-232-509-4334', 5343.69, 'the slyly unusual platelets. always ironic accounts mold blithely. requests use slyly final'),
(6217, 'Supplier#000006217', 'RVN23SYT9jenUeaWGXUd', 3, '13-309-288-5638', 8384.14, 'efully even ideas. express packages boost furiously. carefully fin'),
(6659, 'Supplier#000006659', 'iTLsnvD8D2GzWNUv kRInwRjk5rDeEmfup1', 3, '13-746-779-9180', 1067.93, 'ges boost slyly furiously sly accounts. silent requests are carefully evenly regular'),
(6949, 'Supplier#000006949', 'mLxYUJhsGcLtKe ,GFirNu183AvT', 3, '13-776-634-2532', 2792.37, 'boldly bold gifts are deposits. even accou'),
(7398, 'Supplier#000007398', 'V8eE6oZ00OFNU,', 3, '13-304-352-9285', 4390.85, 't deposits. even, pending instructions wake slyly. blithely regular'),
(7801, 'Supplier#000007801', '69fi,U1r6enUb', 3, '13-235-541-2569', 1296.61, 'lar, even deposits boost. final accounts serve after the furiously pen'),
(7874, 'Supplier#000007874', 'pyghw9iRdkGvkF0Yf5HEqJvmSOohAPFx6U', 3, '13-297-807-1753', 4836.47, 'ets during the bold multipliers are quickly ideas. finally bold foxes poach slowly pending accou'),
(7926, 'Supplier#000007926', 'ErzCF80K9Uy', 3, '13-829-609-6904', 2159.30, 're blithely. even, special braids sleep furiously. deposits boo'),
(8595, 'Supplier#000008595', 'MH0iB73GQ3z UW3O DbCbqmc', 3, '13-841-273-9262', 2195.24, 'ily silent deposits into the slyly silent request'),
(8610, 'Supplier#000008610', 'SgVgP90vP452sUNTgzL9zKwXHXAzV6tV', 3, '13-107-134-7845', 307.37, 'lar theodolites sleep carefully'),
(8879, 'Supplier#000008879', 'rDSA,D9oPM,65NMWEFrmGKAu', 3, '13-477-693-9090', 1393.64, 'ing pinto beans. blithely pending dependencies about the ironic deposits print according to t'),
(9032, 'Supplier#000009032', 'qK,trB6Sdy4Dz1BRUFNy', 3, '13-441-662-5539', 3982.32, 'regular requests use blithely express packages. theodolites cajole boldly along the ideas. package'),
(9430, 'Supplier#000009430', 'igRqmneFt', 3, '13-491-379-4658', 3345.48, 'aggle blithely at the silent warthogs. regular foxes cajole along th'),
(9718, 'Supplier#000009718', 'TJcCKxiE0v8hVqydfBW8EFuRqNDeKoD', 3, '13-817-490-5383', 6861.64, 'onic ideas are blithely carefully pending deposits. deposits are blithely regular pa'),
(9812, 'Supplier#000009812', 'APFRMy3lCbgFga53n5t9DxzFPQPgnjrGt32', 3, '13-298-591-2357', 2709.63, 'y final accounts dazzle quickly. express theodolites cajole carefu'),
(9899, 'Supplier#000009899', '7XdpAHrzr1t,UQFZE', 3, '13-200-446-4635', 4235.99, 'print blithely. carefully final theodolites alongside of the blithely even');