    case IDS_TABLES:  p = "Table Headers";    break;
    case IDS_TAB_01:  p = "Table_Cat";        break;
    case IDS_TAB_02:  p = "Table_Schema";     break;
    case IDS_TAB_03:  p = "Table_Name";       break;
    case IDS_TAB_04:  p = "Table_Type";       break;
    case IDS_TAB_05:  p = "Remark";           break;
    case IDS_COLUMNS: p = "Column Headers";   break;
    case IDS_COL_01:  p = "Table_Cat";        break;
    case IDS_COL_02:  p = "Table_Schema";     break;
    case IDS_COL_03:  p = "Table_Name";       break;
    case IDS_COL_04:  p = "Column_Name";      break;
    case IDS_COL_05:  p = "Data_Type";        break;
    case IDS_COL_06:  p = "Type_Name";        break;
    case IDS_COL_07:  p = "Column_Size";      break;
    case IDS_COL_08:  p = "Buffer_Length";    break;
    case IDS_COL_09:  p = "Decimal_Digits";   break;
    case IDS_COL_10:  p = "Radix";            break;
    case IDS_COL_11:  p = "Nullable";         break;
    case IDS_COL_12:  p = "Remarks";          break;
    case IDS_PKEY:    p = "Key Headers";      break;
    case IDS_PKY_01:  p = "Table_Catalog";    break;
    case IDS_PKY_02:  p = "Table_Schema";     break;
    case IDS_PKY_03:  p = "Table_Name";       break;
    case IDS_PKY_04:  p = "Column_Name";      break;
    case IDS_PKY_05:  p = "Key_Seq";          break;
    case IDS_PKY_06:  p = "Pk_Name";          break;
    case IDS_STAT:    p = "Stat Headers";     break;
    case IDS_STA_01:  p = "Table_Catalog";    break;
    case IDS_STA_02:  p = "Table_Schema";     break;
    case IDS_STA_03:  p = "Table_Name";       break;
    case IDS_STA_04:  p = "Non_Unique";       break;
    case IDS_STA_05:  p = "Index_Qualifier";  break;
    case IDS_STA_06:  p = "Index_Name";       break;
    case IDS_STA_07:  p = "Type";             break;
    case IDS_STA_08:  p = "Seq_in_Index";     break;
    case IDS_STA_09:  p = "Column_Name";      break;
    case IDS_STA_10:  p = "Collation";        break;
    case IDS_STA_11:  p = "Cardinality";      break;
    case IDS_STA_12:  p = "Pages";            break;
    case IDS_STA_13:  p = "Filter_Condition"; break;
    case IDS_DRIVER:  p = "Driver Headers";   break;
    case IDS_DRV_01:  p = "Description";      break;
    case IDS_DRV_02:  p = "Attributes";       break;
    case IDS_DSRC:    p = "DataSrc Headers";  break;
    case IDS_DSC_01:  p = "Name";             break;
    case IDS_DSC_02:  p = "Description";      break;
