CREATE TABLE `part` (
  `P_PARTKEY` int(11) NOT NULL,
  `P_NAME` varchar(55) NOT NULL,
  `P_MFGR` char(25) NOT NULL,
  `P_BRAND` char(10) NOT NULL,
  `P_TYPE` varchar(25) NOT NULL,
  `P_SIZE` int(11) NOT NULL,
  `P_CONTAINER` char(10) NOT NULL,
  `P_RETAILPRICE` decimal(15,2) NOT NULL,
  `P_COMMENT` varchar(23) NOT NULL
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb3 COLLATE=utf8mb3_general_ci;

INSERT INTO `part` (
    `P_PARTKEY`, `P_NAME`, `P_MFGR`, `P_BRAND`, `P_TYPE`, 
    `P_SIZE`, `P_CONTAINER`, `P_RETAILPRICE`, `P_COMMENT`
) VALUES
(939, 'salmon magenta orange coral aquamarine', 'Manufacturer#1', 'Brand#14', 'PROMO BURNISHED BRASS', 41, 'WRAP PACK', 1839.93, 'ages wake doggedly of'),
(966, 'turquoise antique brown violet plum', 'Manufacturer#4', 'Brand#42', 'PROMO BURNISHED TIN', 24, 'MED CAN', 1866.96, ' the carefully'),
(1085, 'floral frosted seashell light red', 'Manufacturer#4', 'Brand#45', 'PROMO PLATED COPPER', 26, 'LG CAN', 986.08, 'anent notornis about'),
(1239, 'purple dodger lime peach burnished', 'Manufacturer#5', 'Brand#51', 'PROMO BURNISHED STEEL', 24, 'LG CASE', 1140.23, 'structions? quickly i'),
(1343, 'navy metallic green cornsilk blue', 'Manufacturer#3', 'Brand#31', 'PROMO BURNISHED TIN', 11, 'WRAP PKG', 1244.34, 'cuses'),
(2988, 'medium white firebrick cream sky', 'Manufacturer#3', 'Brand#33', 'PROMO POLISHED COPPER', 3, 'MED DRUM', 1890.98, '. boldly pendin'),
(3024, 'lemon burnished goldenrod khaki saddle', 'Manufacturer#1', 'Brand#15', 'PROMO BURNISHED COPPER', 10, 'MED PACK', 927.02, ', eve'),
(3126, 'chiffon azure orange seashell blanched', 'Manufacturer#3', 'Brand#34', 'PROMO POLISHED COPPER', 1, 'LG DRUM', 1029.12, 'special dolphins al'),
(3142, 'papaya dim beige salmon pink', 'Manufacturer#3', 'Brand#33', 'PROMO BRUSHED STEEL', 50, 'SM CAN', 1045.14, 'quests. slyly'),
(3357, 'violet dim metallic lemon spring', 'Manufacturer#3', 'Brand#32', 'PROMO ANODIZED COPPER', 40, 'SM CASE', 1260.35, 'lent pl'),
(3625, 'indian peach coral brown linen', 'Manufacturer#5', 'Brand#54', 'PROMO PLATED TIN', 50, 'MED CASE', 1528.62, 'ccounts cajole ca'),
(4119, 'blanched salmon gainsboro grey coral', 'Manufacturer#4', 'Brand#43', 'PROMO BRUSHED BRASS', 40, 'JUMBO PACK', 1023.11, 'dependencies wa'),
(4466, 'black white chocolate blue drab', 'Manufacturer#2', 'Brand#21', 'PROMO ANODIZED NICKEL', 22, 'SM CAN', 1370.46, 'r the regular braids:'),
(88035, 'cornsilk navajo cornflower metallic midnight', 'Manufacturer#4', 'Brand#45', 'PROMO BRUSHED NICKEL', 16, 'JUMBO DRUM', 1023.03, 'ly even theodolites a'),
(88362, 'ivory dodger thistle royal olive', 'Manufacturer#2', 'Brand#21', 'ECONOMY POLISHED STEEL', 39, 'WRAP BAG', 1350.36, 'even requests.'),
(123927, 'firebrick powder light chiffon violet', 'Manufacturer#1', 'Brand#14', 'MEDIUM PLATED TIN', 41, 'WRAP CAN', 1950.92, 'ackages. blithely'),
(128449, 'cream linen salmon smoke blue', 'Manufacturer#2', 'Brand#25', 'STANDARD BURNISHED BRASS', 10, 'WRAP BOX', 1477.44, 'nal, regular inst'),
(137469, 'steel royal frosted peru sienna', 'Manufacturer#4', 'Brand#42', 'ECONOMY BURNISHED COPPER', 17, 'LG PKG', 1506.46, 'unts aft'),
(155190, 'chocolate lace cornflower rosy light', 'Manufacturer#4', 'Brand#44', 'PROMO BRUSHED NICKEL', 9, 'JUMBO JAR', 1245.19, 'sly unusual orbits use'),
(182052, 'blue floral drab yellow dark', 'Manufacturer#5', 'Brand#52', 'MEDIUM PLATED COPPER', 38, 'JUMBO PKG', 1134.05, 'ic, regular foxes. bo'),
(197921, 'goldenrod yellow royal salmon linen', 'Manufacturer#5', 'Brand#53', 'MEDIUM BRUSHED BRASS', 10, 'WRAP DRUM', 2018.92, 'r the quickl'),
(2217, 'forest lavender snow thistle peru', 'Manufacturer#4', 'Brand#41', 'PROMO BURNISHED NICKEL', 47, 'SM PKG', 1119.21, 'regular'),
(4354, 'forest navajo medium salmon mint', 'Manufacturer#4', 'Brand#45', 'PROMO BURNISHED BRASS', 16, 'WRAP PKG', 1258.35, 'deposits are'),
(4668, 'forest lemon metallic blush orange', 'Manufacturer#3', 'Brand#35', 'STANDARD PLATED COPPER', 9, 'LG DRUM', 1572.66, 'es thrash qui'),
(4918, 'forest orange midnight green cornflower', 'Manufacturer#5', 'Brand#53', 'ECONOMY BRUSHED TIN', 45, 'LG BAG', 1822.91, 'egular ideas hagg'),
(6092, 'forest goldenrod deep maroon misty', 'Manufacturer#3', 'Brand#34', 'PROMO ANODIZED STEEL', 22, 'JUMBO JAR', 998.09, 'ns run f'),
(7281, 'forest lime wheat chartreuse violet', 'Manufacturer#4', 'Brand#41', 'PROMO ANODIZED COPPER', 9, 'JUMBO CAN', 1188.28, ' ironic depende'),
(8184, 'forest powder medium saddle bisque', 'Manufacturer#5', 'Brand#54', 'STANDARD POLISHED COPPER', 22, 'MED BAG', 1092.18, 'ven accounts ac'),
(9703, 'forest steel deep lemon goldenrod', 'Manufacturer#3', 'Brand#34', 'STANDARD ANODIZED COPPER', 45, 'SM BAG', 1612.70, 'symptot'),
(10297, 'forest tomato grey linen plum', 'Manufacturer#2', 'Brand#22', 'ECONOMY BRUSHED NICKEL', 33, 'MED PACK', 1207.29, 'equests nag slyly b'),
(10439, 'forest salmon red saddle hot', 'Manufacturer#4', 'Brand#43', 'SMALL PLATED STEEL', 32, 'JUMBO PACK', 1349.43, ', pending theodo'),
(10920, 'forest sandy salmon lace sienna', 'Manufacturer#4', 'Brand#41', 'ECONOMY ANODIZED TIN', 39, 'JUMBO DRUM', 1830.92, 'ly idly close pl'),
(11926, 'forest red peru blush dodger', 'Manufacturer#2', 'Brand#21', 'LARGE BRUSHED TIN', 6, 'SM CASE', 1837.92, 'ironic waters nag'),
(11946, 'forest rose metallic spring peru', 'Manufacturer#2', 'Brand#23', 'STANDARD BRUSHED BRASS', 25, 'SM PKG', 1857.94, 'ccounts.'),
(12395, 'forest steel blue mint gainsboro', 'Manufacturer#4', 'Brand#44', 'MEDIUM BRUSHED STEEL', 44, 'LG PKG', 1307.39, 'fily ex'),
(12924, 'forest royal magenta rose sky', 'Manufacturer#3', 'Brand#31', 'LARGE BRUSHED BRASS', 49, 'WRAP BOX', 1836.92, 'eas unwind quickly c'),
(13592, 'forest dodger orchid gainsboro azure', 'Manufacturer#4', 'Brand#43', 'MEDIUM BRUSHED BRASS', 6, 'WRAP CAN', 1505.59, 'account'),
(13607, 'forest blanched plum white cornsilk', 'Manufacturer#5', 'Brand#54', 'STANDARD ANODIZED STEEL', 2, 'WRAP DRUM', 1520.60, 'y regular'),
(14029, 'forest bisque azure ghost indian', 'Manufacturer#5', 'Brand#52', 'ECONOMY ANODIZED STEEL', 3, 'SM DRUM', 943.02, 'yly bold foxes'),
(15784, 'forest puff azure misty dodger', 'Manufacturer#2', 'Brand#21', 'PROMO BURNISHED COPPER', 33, 'SM JAR', 1699.78, 'ly ironic accounts'),
(16108, 'forest pale floral blue dim', 'Manufacturer#3', 'Brand#34', 'MEDIUM BRUSHED BRASS', 4, 'MED BOX', 1024.10, 'l accounts do boos'),
(2217, 'forest lavender snow thistle peru', 'Manufacturer#4', 'Brand#41', 'PROMO BURNISHED NICKEL', 47, 'SM PKG', 1119.21, 'regular'),
(4668, 'forest lemon metallic blush orange', 'Manufacturer#3', 'Brand#35', 'STANDARD PLATED COPPER', 9, 'LG DRUM', 1572.66, 'es thrash qui'),
(4918, 'forest orange midnight green cornflower', 'Manufacturer#5', 'Brand#53', 'ECONOMY BRUSHED TIN', 45, 'LG BAG', 1822.91, 'egular ideas hagg'),
(6092, 'forest goldenrod deep maroon misty', 'Manufacturer#3', 'Brand#34', 'PROMO ANODIZED STEEL', 22, 'JUMBO JAR', 998.09, 'ns run f'),
(6216, 'forest beige red blush deep', 'Manufacturer#5', 'Brand#51', 'MEDIUM BRUSHED BRASS', 47, 'MED JAR', 1122.21, 'ly about the furiou'),
(6729, 'forest mint cyan goldenrod azure', 'Manufacturer#1', 'Brand#12', 'LARGE BURNISHED TIN', 36, 'LG BOX', 1635.72, 'ong the warthogs i'),
(7281, 'forest lime wheat chartreuse violet', 'Manufacturer#4', 'Brand#41', 'PROMO ANODIZED COPPER', 9, 'JUMBO CAN', 1188.28, ' ironic depende'),
(8184, 'forest powder medium saddle bisque', 'Manufacturer#5', 'Brand#54', 'STANDARD POLISHED COPPER', 22, 'MED BAG', 1092.18, 'ven accounts ac'),
(9703, 'forest steel deep lemon goldenrod', 'Manufacturer#3', 'Brand#34', 'STANDARD ANODIZED COPPER', 45, 'SM BAG', 1612.70, 'symptot'),
(10297, 'forest tomato grey linen plum', 'Manufacturer#2', 'Brand#22', 'ECONOMY BRUSHED NICKEL', 33, 'MED PACK', 1207.29, 'equests nag slyly b'),
(10370, 'forest orchid deep antique peru', 'Manufacturer#2', 'Brand#22', 'SMALL PLATED TIN', 9, 'WRAP CASE', 1280.37, 'final instruction'),
(10439, 'forest salmon red saddle hot', 'Manufacturer#4', 'Brand#43', 'SMALL PLATED STEEL', 32, 'JUMBO PACK', 1349.43, ', pending theodo'),
(10920, 'forest sandy salmon lace sienna', 'Manufacturer#4', 'Brand#41', 'ECONOMY ANODIZED TIN', 39, 'JUMBO DRUM', 1830.92, 'ly idly close pl'),
(11244, 'forest pink snow saddle sienna', 'Manufacturer#3', 'Brand#33', 'PROMO BRUSHED NICKEL', 22, 'JUMBO DRUM', 1155.24, 'ep slyly deposits.'),
(11512, 'forest light peru coral thistle', 'Manufacturer#5', 'Brand#53', 'SMALL BRUSHED STEEL', 21, 'MED PKG', 1423.51, 'inal forges. carefull'),
(11926, 'forest red peru blush dodger', 'Manufacturer#2', 'Brand#21', 'LARGE BRUSHED TIN', 6, 'SM CASE', 1837.92, 'ironic waters nag'),
(11946, 'forest rose metallic spring peru', 'Manufacturer#2', 'Brand#23', 'STANDARD BRUSHED BRASS', 25, 'SM PKG', 1857.94, 'ccounts.'),
(12020, 'forest lavender mint navy blush', 'Manufacturer#3', 'Brand#35', 'PROMO BURNISHED STEEL', 21, 'WRAP CASE', 932.02, 'er th'),
(12395, 'forest steel blue mint gainsboro', 'Manufacturer#4', 'Brand#44', 'MEDIUM BRUSHED STEEL', 44, 'LG PKG', 1307.39, 'fily ex'),
(12924, 'forest royal magenta rose sky', 'Manufacturer#3', 'Brand#31', 'LARGE BRUSHED BRASS', 49, 'WRAP BOX', 1836.92, 'eas unwind quickly c'),
(13379, 'forest moccasin gainsboro slate lemon', 'Manufacturer#2', 'Brand#25', 'MEDIUM ANODIZED BRASS', 39, 'MED JAR', 1292.37, 'refully final pint'),
(13592, 'forest dodger orchid gainsboro azure', 'Manufacturer#4', 'Brand#43', 'MEDIUM BRUSHED BRASS', 6, 'WRAP CAN', 1505.59, 'account'),
(13607, 'forest blanched plum white cornsilk', 'Manufacturer#5', 'Brand#54', 'STANDARD ANODIZED STEEL', 2, 'WRAP DRUM', 1520.60, 'y regular'),
(14029, 'forest bisque azure ghost indian', 'Manufacturer#5', 'Brand#52', 'ECONOMY ANODIZED STEEL', 3, 'SM DRUM', 943.02, 'yly bold foxes'),
(14200, 'forest thistle black grey ghost', 'Manufacturer#3', 'Brand#31', 'PROMO PLATED STEEL', 1, 'MED DRUM', 1114.20, 'each blithe'),
(15784, 'forest puff azure misty dodger', 'Manufacturer#2', 'Brand#21', 'PROMO BURNISHED COPPER', 33, 'SM JAR', 1699.78, 'ly ironic accounts'),
(16024, 'forest purple navajo lavender misty', 'Manufacturer#3', 'Brand#35', 'ECONOMY ANODIZED BRASS', 16, 'LG JAR', 940.02, 'xes sle'),
(16108, 'forest pale floral blue dim', 'Manufacturer#3', 'Brand#34', 'MEDIUM BRUSHED BRASS', 4, 'MED BOX', 1024.10, 'l accounts do boos'),
(16377, 'forest moccasin ghost blanched olive', 'Manufacturer#5', 'Brand#55', 'PROMO BRUSHED STEEL', 37, 'WRAP CASE', 1293.37, ' blithely fin'),
(16697, 'forest tomato misty rose dim', 'Manufacturer#4', 'Brand#45', 'ECONOMY PLATED NICKEL', 9, 'MED BAG', 1613.69, ' haggle blith'),
(17300, 'forest sky light violet chartreuse', 'Manufacturer#3', 'Brand#31', 'MEDIUM BRUSHED STEEL', 1, 'WRAP PKG', 1217.30, 'ickly even'),
(17310, 'forest dim royal sandy plum', 'Manufacturer#3', 'Brand#32', 'MEDIUM ANODIZED STEEL', 8, 'JUMBO BOX', 1227.31, 'instructions wake. b'),
(17692, 'forest burlywood tomato drab rosy', 'Manufacturer#5', 'Brand#55', 'SMALL BURNISHED TIN', 3, 'LG JAR', 1609.69, 'fluff'),
(17925, 'forest cream blush turquoise saddle', 'Manufacturer#2', 'Brand#21', 'STANDARD POLISHED COPPER', 24, 'MED BAG', 1842.92, ' even pinto beans da'),
(18227, 'forest lavender orchid blush mint', 'Manufacturer#2', 'Brand#22', 'ECONOMY BURNISHED TIN', 35, 'LG BAG', 1145.22, 'l courts. quickly ir'),
(18370, 'forest rose smoke salmon slate', 'Manufacturer#1', 'Brand#12', 'MEDIUM BURNISHED BRASS', 3, 'WRAP PKG', 1288.37, 'e carefully'),
(20294, 'forest lace seashell purple blanched', 'Manufacturer#3', 'Brand#33', 'ECONOMY BURNISHED BRASS', 43, 'JUMBO CASE', 1214.29, 'owly r'),
(20421, 'forest chartreuse pink azure indian', 'Manufacturer#4', 'Brand#41', 'MEDIUM BURNISHED TIN', 43, 'SM JAR', 1341.42, 'ular r'),
(21654, 'forest brown lace snow honeydew', 'Manufacturer#4', 'Brand#44', 'LARGE BRUSHED COPPER', 32, 'SM PACK', 1575.65, 's nag'),
(21801, 'forest thistle light lemon gainsboro', 'Manufacturer#4', 'Brand#42', 'ECONOMY PLATED COPPER', 24, 'LG JAR', 1722.80, 'ncies w'),
(23023, 'forest indian turquoise medium pink', 'Manufacturer#2', 'Brand#21', 'LARGE BURNISHED TIN', 14, 'MED CAN', 946.02, 'ly. even, pendi'),
(23200, 'forest dodger hot tan cornflower', 'Manufacturer#4', 'Brand#41', 'SMALL BURNISHED TIN', 17, 'JUMBO PACK', 1123.20, 'ackages.'),
(25122, 'forest pale orange cyan red', 'Manufacturer#4', 'Brand#45', 'LARGE PLATED TIN', 31, 'MED DRUM', 1047.12, 'ill cajole a');