/**
 * Autogenerated by Thrift Compiler (0.9.1)
 *
 * DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
 *  @generated
 */
#pragma once

#include "querytele_types.h"

namespace querytele
{
class queryteleConstants
{
 public:
  queryteleConstants();
};

extern const queryteleConstants g_querytele_constants;

}  // namespace querytele
