CREATE INDEX calpont.index_name ON calpont.tbl_name(col_name);
CREATE INDEX calpont.index_name ON calpont.tbl_name(col_name(10));
CREATE INDEX index_name ON tbl_name(col_name,col_name);
CREATE INDEX calpont.index_name ON calpont.tbl_name(col_name,col_name) engine=infinidb;
