DROP TABLE tbl_name;
DROP TABLE tbl_name CASCADE CONSTRAINTS;
DROP TABLE tbl_name engine=infinidb;
DROP TABLE tbl_name CASCADE CONSTRAINTS engine=infinidb;
drop table tbl_name purge;
DROP TABLE tbl_name,tbl_name2;
