.slideshow {
  position: relative;

  .slide {
    display: none;

    img {
      display: block;
      margin: 0 auto;
    }

    &.slideActive {
      display: block;
    }

    a {
      border: none;
      display: block;
    }
  }

  .pagination {
    display: block;
    margin: -10px;
    padding: 1em 0;
    text-align: center;
    width: 100%;

    .pager {
      background: transparent;
      border: 2px solid rgba(255, 255, 255, 0.5);
      border-radius: 50%;
      cursor: pointer;
      display: inline-block;
      height: 12px;
      margin: 10px;
      transition: background-color 0.3s, border-color 0.3s;
      width: 12px;

      &.pagerActive {
        background: rgba(255, 255, 255, 0.5);
        border-width: 4px;
        height: 8px;
        width: 8px;
      }
    }
  }
}
