alter table lineitem add column CBIGINT BIGINT;
alter table lineitem add column CDECIMAL1 DECIMAL(1);
alter table lineitem add column CDECIMAL4 DECIMAL(4);
alter table lineitem add column CDECIMAL4_2 DECIMAL(4,2);
alter table lineitem add column CDECIMAL5 DECIMAL(5);
alter table lineitem add column CDECIMAL9 DECIMAL(9);
alter table lineitem add column CDECIMAL9_2 DECIMAL(9,2);
alter table lineitem add column CDECIMAL10 DECIMAL(10);
alter table lineitem add column CDECIMAL18 DECIMAL(18);
alter table lineitem add column CDECIMAL18_2 DECIMAL(18,2);
alter table lineitem add column CINTEGER INTEGER;
alter table lineitem add column CSMALLINT SMALLINT;
alter table lineitem add column CTINYINT TINYINT;
alter table lineitem add column CDOUBLE DOUBLE;
alter table lineitem add column CFLOAT FLOAT;
alter table lineitem add column CDATE DATE;
alter table lineitem add column CDATETIME DATETIME;
alter table lineitem add column CTIME TIME;
alter table lineitem add column CCHAR1 CHAR(1);
alter table lineitem add column CCHAR2 CHAR(2);
alter table lineitem add column CCHAR3 CHAR(3);
alter table lineitem add column CCHAR4 CHAR(4);
alter table lineitem add column CCHAR5 CHAR(5);
alter table lineitem add column CCHAR6 CHAR(6);
alter table lineitem add column CCHAR7 CHAR(7);
alter table lineitem add column CCHAR8 CHAR(8);
alter table lineitem add column CCHAR9 CHAR(9);
alter table lineitem add column CCHAR255 CHAR(255);
alter table lineitem add column CVCHAR1 VARCHAR(1);
alter table lineitem add column CVCHAR2 VARCHAR(2);
alter table lineitem add column CVCHAR3 VARCHAR(3);
alter table lineitem add column CVCHAR4 VARCHAR(4);
alter table lineitem add column CVCHAR5 VARCHAR(5);
alter table lineitem add column CVCHAR6 VARCHAR(6);
alter table lineitem add column CVCHAR7 VARCHAR(7);
alter table lineitem add column CVCHAR8 VARCHAR(8);
alter table lineitem add column CVCHAR255 VARCHAR(255);
alter table lineitem add column CTEXT TEXT;
alter table lineitem add column CTINYTEXT TINYTEXT;
alter table lineitem add column CMEDIUMTEXT MEDIUMTEXT;
alter table lineitem add column CLONGTEXT LONGTEXT;
alter table lineitem add column CBLOB BLOB;
alter table lineitem add column CTINYBLOB TINYBLOB;
alter table lineitem add column CMEDIUMBLOB MEDIUMBLOB;
alter table lineitem add column CLONGBLOB LONGBLOB;
