select * from t1 where id=1;
select * from t1 where value=1;
select value from t1 where value=1;
select * from t1 where value2=1;
select * from t1 where id=5;
select * from t1 where value=5;
select value from t1 where value=5;
select * from t1 where value2=5;
select * from t1 where id < 3 order by id;
select * from t1 where value < 3 order by id;
select value from t1 where value < 3 order by id;
select * from t1 where value2 < 3 order by id;
select * from t1 order by id;
select value from t1 order by id;
