LIBRARY "wixca"
VERSION 1.0
EXPORTS
PresetDatabaseProperties
RemoveDataDirectory
CreateDatabaseRollback
CheckDatabaseProperties
CheckDataDirectory
CheckDBInUse
CheckServiceUpgrades
CheckInstallDirectory
